use ctap::{get_assertion_devices, FidoAssertionRequestBuilder, FidoDevice, FidoErrorKind};
use pamsm::PamLibExt;
use pamsm::{Pam, PamError, PamFlag, PamServiceModule};
use rand::Rng;
use regex::Regex;
use std::collections::HashMap;
use std::error::Error;
use std::fs::File;
use std::io::{self, prelude::*, BufReader};
use std::path::Path;
use std::time::{Duration, SystemTime};

use failure::ResultExt;

pub struct PamFido2;

struct Settings {
    pub device_timeout: Duration,
    pub user_credentials: Vec<(String, String)>,
    pub debug: bool,
}

impl Default for Settings {
    fn default() -> Self {
        Settings {
            debug: false,
            device_timeout: Duration::from_secs(15),
            user_credentials: vec![("091566e43802c5a29971c1e08d7865d959af862cc28af22dacf413ac26b90f6dea7d1ac491d9d3712c63f7b8d6cfadf86d057d099d382246dbe9c87f133ed167881b65030000".into(),".*".into())],
        }
    }
}

impl Settings {
    pub fn from_args(args: Vec<String>) -> Result<Self, Box<dyn Error>> {
        let args = args
            .into_iter()
            .filter_map(|arg| {
                let mut parts = arg.split("=");
                parts
                    .by_ref()
                    .next()
                    .map(|key| (key.to_string(), parts.collect::<Vec<_>>().join("=")))
            })
            .collect::<HashMap<String, String>>();
        let mut settings = Settings::load(
            args.get("authfile")
                .map(AsRef::as_ref)
                .unwrap_or("/etc/fido2_pam.conf"),
        )?;
        let timeout = args
            .get("timeout")
            .map(|t| {
                t.parse::<u64>()
                    .expect("invalid config timeout is supposed to be an int")
            })
            .map(Duration::from_secs);
        if let Some(timeout) = timeout {
            settings.device_timeout = timeout;
        };
        if let Some(debug) = args.get("debug").and_then(|d| d.parse::<bool>().ok()) {
            settings.debug = debug;
        }
        Ok(settings)
    }
    pub fn load(path: impl AsRef<Path>) -> io::Result<Settings> {
        let mut creds = Vec::new();
        let file = File::open(path.as_ref())?;
        let reader = BufReader::new(file);

        for line in reader.lines() {
            let line = line?;
            let mut parts = line.split(":");
            if let Some(user) = parts.by_ref().next() {
                creds.push((
                    (&parts.collect::<Vec<_>>()[..].join(":")).to_string(),
                    user.to_string(),
                ));
            }
        }

        Ok(Settings {
            user_credentials: creds,
            ..Default::default()
        })
    }

    pub fn get_credentials(&self, user: &str) -> Vec<ctap::FidoCredential> {
        let mut creds = Vec::new();
        for (cred, pattern) in self.user_credentials.iter() {
            let re = Regex::new(&pattern).expect(&["Invalid regex pattern:", &pattern].join(" "));
            if re.is_match(user) {
                let mut parts = cred.split(":");
                match (parts.by_ref().next(), parts.by_ref().next()) {
                    (Some(id), Some(key)) => {
                        creds.push(ctap::FidoCredential {
                            id: hex::decode(id).unwrap(),
                            public_key: hex::decode(key).ok(),
                        });
                    }
                    _ => eprintln!("Failed to parse {}", cred),
                }
            }
        }
        creds
    }
}

impl PamFido2 {
    pub fn fido_authenticate(
        &self,
        username: &str,
        args: Vec<String>,
    ) -> Result<PamError, Box<dyn Error>> {
        let settings = Settings::from_args(args)?;
        let debug = |msg: &dyn Fn() -> String| {
            if settings.debug {
                eprintln!("{}", msg())
            }
        };
        let begin = SystemTime::now();
        let challenge = rand::thread_rng().gen::<[u8; 32]>();
        let credentials = settings.get_credentials(username);
        if credentials.is_empty() {
            debug(&|| format!("Found no credentials for {}", username));
            return Ok(PamError::AUTH_ERR);
        }
        debug(&|| format!("Found {} credentials for {}", credentials.len(), username));
        let slice = credentials.iter().collect::<Vec<_>>();
        let request = FidoAssertionRequestBuilder::default()
            .rp_id("fido2pam")
            .client_data_hash(&challenge[..])
            .credentials(&slice[..])
            .build()
            .unwrap();

        loop {
            let mut devices = ctap::get_devices()
                .compat()?
                .filter_map(|handle| FidoDevice::new(&handle).ok())
                .collect::<Vec<_>>();
            if begin.elapsed().unwrap() > settings.device_timeout {
                debug(&|| format!("Exceeded {:?} timeout", settings.device_timeout));
                return Ok(PamError::AUTH_ERR);
            }
            if devices.is_empty() {
                debug(&|| "Please connect your authenicator!".into());
                std::thread::sleep(Duration::from_secs(1));
                continue;
            }
            match get_assertion_devices(
                &request,
                devices.iter_mut(),
                Some(settings.device_timeout / 8u32),
            ) {
                Ok((cred, _)) => {
                    debug(&|| format!("Got assertion for credential: {}", hex::encode(&cred.id)));
                    return Ok(PamError::SUCCESS);
                }
                Err(_) if begin.elapsed().unwrap() > settings.device_timeout => {
                    debug(&|| format!("Exceeded {:?} timeout", settings.device_timeout));
                    return Ok(PamError::AUTH_ERR);
                }
                Err(e)
                    if e.kind() == FidoErrorKind::DeviceUnsupported
                        || e.kind() == FidoErrorKind::Timeout =>
                {
                    continue
                }
                Err(e) => Err(e).compat()?,
            }
        }
    }
}

impl PamServiceModule for PamFido2 {
    fn authenticate(pamh: Pam, _: PamFlag, args: Vec<String>) -> PamError {
        let username = &pamh
            .get_cached_user()
            .ok()
            .map(|name| name.unwrap().to_str().unwrap().to_string())
            .expect("Failed to get username");

        match PamFido2.fido_authenticate(username, args) {
            Ok(e) => e,
            Err(e) => {
                eprintln!("{}", e);
                PamError::AUTH_ERR
            }
        }
    }
}
